(function (global, factory) {
    typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
    typeof define === 'function' && define.amd ? define(['exports'], factory) :
    (global = typeof globalThis !== 'undefined' ? globalThis : global || self, factory(global["Stimulus Typescript"] = {}));
})(this, (function (exports) { 'use strict';

    /******************************************************************************
    Copyright (c) Microsoft Corporation.

    Permission to use, copy, modify, and/or distribute this software for any
    purpose with or without fee is hereby granted.

    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
    REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
    AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
    INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
    LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
    OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
    PERFORMANCE OF THIS SOFTWARE.
    ***************************************************************************** */

    function __setFunctionName(f, name, prefix) {
        if (typeof name === "symbol") name = name.description ? "[".concat(name.description, "]") : "";
        return Object.defineProperty(f, "name", { configurable: true, value: prefix ? "".concat(prefix, " ", name) : name });
    }
    typeof SuppressedError === "function" ? SuppressedError : function (error, suppressed, message) {
        var e = new Error(message);
        return e.name = "SuppressedError", e.error = error, e.suppressed = suppressed, e;
    };

    class Wrapped {
        constructor() {
            // @ts-ignore
            this._ = undefined;
        }
    }
    /**
     * Strongly type Object values
     * ```ts
     * const values = {
     *  address: Object_<{ street: string }>
     * }
     * ```
     */
    const Object_ = Wrapped;
    const ObjectAs = Object_;
    /**
     * Strongly type custom targets
     * ```ts
     * const targets = {
     *  select: Target<CustomSelect>
     * }
     * ```
     */
    const Target = Wrapped;
    /**
     * Convert typed Object_ to ObjectConstructor before passing values to Stimulus
     */
    function patchValueTypeDefinitionMap(values) {
        const patchObject = (def) => {
            if ("type" in def) {
                return {
                    type: def.type === Object_ ? Object : def.type,
                    default: def.default,
                };
            }
            else {
                return def === Object_ ? Object : def;
            }
        };
        return Object.entries(values).reduce((result, [key, def]) => {
            result[key] = patchObject(def);
            return result;
        }, {});
    }
    /**
     * Strongly typed Controller!
     * ```ts
     * const values = {
     *  name: String,
     *  alias: Array<string>,
     *  address: Object_<{ street: string }>
     * }
     * const targets = { form: HTMLFormElement, "select": Target<CustomSelect> }
     * const outlets = { "user-status": UserStatusController }
     *
     * class MyController extends Typed(Controller, { values, targets, outlets }) {
     *  // Look Ma, no "declare ..."
     *  this.nameValue.split(' ')
     *  this.aliasValue.map(alias => alias.toUpperCase())
     *  this.addressValue.street
     *  this.formTarget.submit()
     *  this.selectTarget.search = "stimulus";
     *  this.userStatusOutlets.forEach(status => status.markAsSelected(event))
     * }
     * ```
     */
    function Typed(Base, statics) {
        var _a;
        const { values, targets, outlets } = statics;
        const derived = (_a = class extends Base {
            },
            __setFunctionName(_a, "derived"),
            _a.values = patchValueTypeDefinitionMap(values !== null && values !== void 0 ? values : {}),
            _a.targets = Object.getOwnPropertyNames(targets !== null && targets !== void 0 ? targets : {}),
            _a.outlets = Object.getOwnPropertyNames(outlets !== null && outlets !== void 0 ? outlets : {}),
            _a);
        return derived;
    }

    exports.ObjectAs = ObjectAs;
    exports.Object_ = Object_;
    exports.Target = Target;
    exports.Typed = Typed;

}));
